@@ -14,8 +14,9 @@ use vars qw(@ISA @EXPORT $VERSION %SubTests @SkipReasons);
 @EXPORT = qw(ok skip sok plan have have_lwp have_http11
              have_cgi have_access have_auth have_module have_apache
              have_min_apache_version have_apache_version have_perl 
-             have_threads under_construction);
-$VERSION = '1.0';
+             have_min_perl_version have_min_module_version
+             have_threads under_construction have_apache_mpm);
+$VERSION = '1.04';
 
 %SubTests = ();
 @SkipReasons = ();
@@ -45,7 +46,14 @@ sub sok (&;$) {
         return;
     }
 
-    ok $sub->();
+    my($package, $filename, $line) = caller;
+
+    # trick ok() into reporting the caller filename/line when a
+    # sub-test fails in sok()
+    return eval <<EOE;
+#line $line $filename
+    ok(\$sub->());
+EOE
 }
 
 #so Perl's Test.pm can be run inside mod_perl
@@ -156,6 +164,9 @@ sub have {
                 $have_all = 0;
             }
         }
+        elsif ($cond =~ /^(0|1)$/) {
+            $have_all = 0 if $cond == 0;
+        }
         else {
             $have_all = 0 unless have_module($cond);
         }
@@ -198,6 +209,29 @@ sub have_module {
     }
 }
 
+sub have_min_perl_version {
+    my $version = shift;
+
+    return 1 if $] >= $version;
+
+    push @SkipReasons, "perl >= $version is required";
+    return 0;
+}
+
+# currently supports only perl modules
+sub have_min_module_version {
+    my($module, $version) = @_;
+
+    # have_module requires the perl module
+    return 0 unless have_module($module);
+
+    my $has_version = $module->VERSION || 0;
+    return 1 if $has_version >= $version;
+
+    push @SkipReasons, "$module version $version or higher is required";
+    return 0;
+}
+
 sub have_cgi {
     have_module('cgi') || have_module('cgid');
 }
@@ -257,6 +291,22 @@ sub have_apache_version {
     }
 }
 
+sub have_apache_mpm {
+    my $wanted = shift;
+    my $cfg = Apache::Test::config();
+    my $current = $cfg->{server}->{mpm};
+
+    if ($current ne $wanted) {
+        push @SkipReasons,
+          "apache $wanted mpm is required," .
+          " this is the $current mpm";
+        return 0;
+    }
+    else {
+        return 1;
+    }
+}
+
 sub config_enabled {
     my $key = shift;
     defined $Config{$key} and $Config{$key} eq 'define';
@@ -404,10 +454,10 @@ But this won't hint the reason for skipping therefore it's better to
 use have():
 
   plan tests => 5,
-      have 'LWP',
-           { "perl >= 5.7.3 is required" => sub { $] >= 5.007003 } };
+      have 'LWP', 
+           { "not Win32" => sub { $^O eq 'MSWin32'} };
 
-see have() for more info.
+see C<have()> for more info.
 
 =item * an C<ARRAY> reference
 
@@ -516,6 +566,16 @@ For example:
 
 requires Apache 2.0.40.
 
+=item have_apache_mpm
+
+Used to require a specific Apache Multi-Processing Module.
+
+For example:
+
+  plan tests => 5, have_apache_mpm('prefork');
+
+requires the prefork MPM.
+
 =item have_perl
 
   plan tests => 5, have_perl 'iolayers';
@@ -529,6 +589,16 @@ whether:
 
   $Config{useithread} eq 'define';
 
+=item have_min_perl_version
+
+Used to require a minimum version of Perl.
+
+For example:
+
+  plan tests => 5, have_min_perl_version("5.008001");
+
+requires Perl 5.8.1 or higher.
+
 =item have_module
 
   plan tests => 5, have_module 'CGI';
@@ -541,6 +611,18 @@ arguments or a reference to a list.
 In case of C modules, depending on how the module name was passed it
 may pass through the following completions:
 
+=item have_min_module_version
+
+Used to require a minimum version of a module
+
+For example:
+
+  plan tests => 5, have_min_module_version(CGI => 2.81);
+
+requires C<CGI.pm> version 2.81 or higher.
+
+Currently works only for perl modules.
+
 =over
 
 =item 1 have_module 'proxy_http.c'
@@ -564,7 +646,8 @@ lookup, turning it into I<'mod_cgi.c'>.
 
   plan tests => 5,
       have 'LWP',
-           { "perl >= 5.8.0 is required" => ($] >= 5.008)          },
+           { "perl >= 5.8.0 and w/ithreads is required" => 
+             ($Config{useperlio} && $] >= 5.008) },
            { "not Win32"                 => sub { $^O eq 'MSWin32' },
              "foo is disabled"           => \&is_foo_enabled,
            },
@@ -578,9 +661,17 @@ scalars, which are passed to have_module(), and hash references. If
 hash references are used, the keys, are strings, containing a reason
 for a failure to satisfy this particular entry, the valuees are the
 condition, which are satisfaction if they return true. If the value is
-a scalar it's used as is. If the value is a code reference, it gets
-executed at the time of check and its return value is used to check
-the condition. If the condition check fails, the provided (in a key)
+0 or 1, it used to decide whether the requirements very satisfied, so
+you can mix special C<have_*()> functions that return 0 or 1. For
+example:
+
+  plan tests => 1, have 'Compress::Zlib', 'deflate',
+      have_min_apache_version("2.0.49");
+
+If the scalar value is a string, different from 0 or 1, it's passed to
+I<have_module()>.  If the value is a code reference, it gets executed
+at the time of check and its return value is used to check the
+condition. If the condition check fails, the provided (in a key)
 reason is used to tell user why the test was skipped.
 
 In the presented example, we require the presense of the C<LWP> Perl
@@ -609,4 +700,24 @@ into a string.  Example:
     # $tests will contain the Test.pm output: 1..4\nok 1\n...
     my $tests = Apache::TestToString->finish;
 
+=head1 SEE ALSO
+
+The Apache-Test tutorial:
+L<http://perl.apache.org/docs/general/testing/testing.html>.
+
+L<Apache::TestRequest|Apache::TestRequest> subclasses LWP::UserAgent and
+exports a number of useful functions for sending request to the Apache test
+server. You can then test the results of those requests.
+
+Use L<Apache::TestMM|Apache::TestMM> in your F<Makefile.PL> to set up your
+distribution for testing.
+
+=head1 AUTHOR
+
+Doug MacEachern with contributions from Geoffrey Young, Philippe
+M. Chiasson, Stas Bekman and others.
+
+Questions can be asked at the test-dev <at> httpd.apache.org list
+For more information see: http://httpd.apache.org/test/.
+
 =cut
