@@ -20,6 +20,7 @@ use constant STARTUP_TIMEOUT => 300; # secs (good for extreme debug cases)
 use subs qw(exit_shell exit_perl);
 
 my %core_files  = ();
+my %original_t_perms = ();
 
 my @std_run      = qw(start-httpd run-tests stop-httpd);
 my @others       = qw(verbose configure clean help ssl http11);
@@ -40,7 +41,7 @@ my %usage = (
    'times=N'         => 'repeat the tests N times',
    'order=mode'      => 'run the tests in one of the modes: (repeat|rotate|random|SEED)',
    'stop-httpd'      => 'stop the test server',
-   'verbose'         => 'verbose output',
+   'verbose[=1]'     => 'verbose output',
    'configure'       => 'force regeneration of httpd.conf (tests will not be run)',
    'clean'           => 'remove all generated test files',
    'help'            => 'display this message',
@@ -158,6 +159,12 @@ sub getopts {
     local *ARGV = $argv;
     my(%opts, %vopts, %conf_opts);
 
+    # a workaround to support -verbose and -verbose=0|1
+    # $Getopt::Long::VERSION > 2.26 can use the "verbose:1" rule
+    # but we have to support older versions as well
+    @ARGV = grep defined, 
+        map {/-verbose=(\d)/ ? ($1 ? '-verbose' : undef) : $_ } @ARGV;
+
     # permute      : optional values can come before the options
     # pass_through : all unknown things are to be left in @ARGV
     Getopt::Long::Configure(qw(pass_through permute));
@@ -214,6 +221,9 @@ sub getopts {
         my %levels = map {$_ => 1} @Apache::TestTrace::Levels;
         if (exists $levels{ $opts{trace} }) {
             $Apache::TestTrace::Level = $opts{trace};
+            # propogate the override for the server-side.
+            # -trace overrides any previous APACHE_TEST_TRACE_LEVEL settings
+            $ENV{APACHE_TEST_TRACE_LEVEL} = $opts{trace};
         }
         else {
             error "unknown trace level: $opts{trace}",
@@ -431,7 +441,8 @@ sub start {
         error "no test server configured, please specify an httpd or ".
               ($test_config->{APXS} ?
                "an apxs other than $test_config->{APXS}" : "apxs").
-               " or put either in your PATH";
+               " or put either in your PATH. For example:\n" .
+               "$0 -httpd /path/to/bin/httpd";
         exit_perl 0;
     }
 
@@ -451,6 +462,8 @@ sub start {
         }
     }
 
+    $self->adjust_t_perms();
+
     if ($opts->{'start-httpd'}) {
         exit_perl 0 unless $server->start;
     }
@@ -489,6 +502,8 @@ sub run_tests {
 sub stop {
     my $self = shift;
 
+    $self->restore_t_perms;
+
     return $self->{server}->stop if $self->{opts}->{'stop-httpd'};
 }
 
@@ -549,8 +564,13 @@ sub run {
 
     $self->getopts(\@argv);
 
+    $self->pre_configure() if $self->can('pre_configure');
+
     $self->{test_config} = $self->new_test_config;
 
+    # make it easy to move the whole distro
+    $self->refresh unless -e $self->{test_config}->{vars}->{top_dir};
+
     $self->warn_core();
 
     $self->{server} = $self->{test_config}->server;
@@ -604,6 +624,42 @@ sub oh {
 #e.g. t/core or t/core.12499
 my $core_pat = '^core(\.\d+)?' . "\$";
 
+# normally would be called after each test
+# and since it updates the list of seen core files
+# scan_core() won't report these again
+# currently used in Apache::TestSmoke
+sub scan_core_incremental {
+    my $self = shift;
+    my $vars = $self->{test_config}->{vars};
+    my $times = 0;
+    my @msg = ();
+
+    finddepth(sub {
+        return unless -f $_;
+        return unless /$core_pat/o;
+        my $core = $File::Find::name;
+        unless (exists $core_files{$core} && $core_files{$core} == -M $core) {
+            # new core file!
+
+            # XXX: could rename the file if it doesn't include the pid
+            # in its name (i.e., just called 'core', instead of 'core.365')
+
+            # XXX: could pass the test name and rename the core file
+            # to use that name as a suffix, plus pid, time or some
+            # other unique identifier, in case the same test is run
+            # more than once and each time it caused a segfault
+            $core_files{$core} = -M $core;
+            my $oh = oh();
+            my $again = $times++ ? "again" : "";
+            push @msg, "oh $oh, server dumped core $again",
+                "for stacktrace, run: gdb $vars->{httpd} -core $core";
+        }
+    }, $vars->{top_dir});
+
+    return @msg;
+
+}
+
 sub scan_core {
     my $self = shift;
     my $vars = $self->{test_config}->{vars};
@@ -612,7 +668,7 @@ sub scan_core {
     finddepth(sub {
         return unless -f $_;
         return unless /$core_pat/o;
-        my $core = "$File::Find::dir/$_";
+        my $core = $File::Find::name;
         if (exists $core_files{$core} && $core_files{$core} == -M $core) {
             # we have seen this core file before the start of the test
             info "an old core file has been found: $core";
@@ -645,6 +701,95 @@ sub warn_core {
     }, $vars->{top_dir});
 }
 
+# this function handles the cases when the test suite is run under
+# 'root':
+#
+# 1. When user 'bar' is chosen to run Apache with, files and dirs
+#    created by 'root' might be not writable/readable by 'bar'
+#
+# 2. when the source is extracted as user 'foo', and the chosen user
+#    to run Apache under is 'bar', in which case normally 'bar' won't
+#    have the right permissions to write into the fs created by 'foo'.
+#
+# We solve that by 'chown -R bar.bar t/' in a portable way.
+#
+# 3. If the parent directory is not rwx for the chosen user, that user
+#    won't be able to read/write the DocumentRoot. In which case we
+#    have nothing else to do, but to tell the user to fix the situation.
+#
+sub adjust_t_perms {
+    my $self = shift;
+
+    return if Apache::TestConfig::WINFU;
+
+    %original_t_perms = (); # reset global
+
+    my $user = getpwuid($>) || '';
+    if ($user eq 'root') {
+        my $vars = $self->{test_config}->{vars};
+        my $user = $vars->{user};
+        my($uid, $gid) = (getpwnam($user))[2..3]
+            or die "Can't find out uid/gid of '$user'";
+
+        warning "root mode: ". 
+            "changing the files ownership to '$user' ($uid:$gid)";
+        finddepth(sub {
+            $original_t_perms{$File::Find::name} = [(stat $_)[4..5]];
+            chown $uid, $gid, $_;
+        }, $vars->{t_dir});
+
+        $self->check_perms($user, $uid, $gid);
+
+    }
+}
+
+sub restore_t_perms {
+    my $self = shift;
+
+    return if Apache::TestConfig::WINFU;
+
+    if (%original_t_perms) {
+        warning "root mode: restoring the original files ownership";
+        my $vars = $self->{test_config}->{vars};
+        while (my($file, $ids) = each %original_t_perms) {
+            next unless -e $file; # files could be deleted
+            chown @$ids, $file;
+        }
+    }
+}
+
+sub check_perms {
+    my ($self, $user, $uid, $gid) = @_;
+
+    # test that the base dir is rwx by the selected non-root user
+    my $vars = $self->{test_config}->{vars};
+    my $dir  = $vars->{t_dir};
+    my $perl = $vars->{perl};
+    my $check = qq[sudo -u '#$uid' $perl -e ] . 
+        qq['print -r "$dir" &&  -w _ && -x _ ? "OK" : "NOK"'];
+    warning "$check\n";
+    my $res   = qx[$check] || '';
+    warning "result: $res";
+    unless ($res eq 'OK') {
+        #$self->restore_t_perms;
+        error(<<"EOI") && die "\n";
+You are running the test suite under user 'root'.
+Apache cannot spawn child processes as 'root', therefore
+we attempt to run the test suite with user '$user' ($uid:$gid).
+The problem is that the path:
+  $dir
+must be 'rwx' by user '$user', so Apache can read and write under that
+path.
+
+There several ways to resolve this issue. For example move 
+'$dir' to '/tmp/' and repeat the 'make test' phase. 
+
+You can test whether the location is good by running the following test:
+  % $check
+EOI
+    }
+}
+
 sub run_request {
     my($test_config, $opts) = @_;
 
@@ -762,7 +907,7 @@ usage: TEST [options ...]
 EOM
 
     for (sort keys %usage){
-        printf "   -%-16s %s\n", $_, $usage{$_};
+        printf "  -%-13s %s\n", $_, $usage{$_};
     }
 
     print "\n   configuration options:\n";
@@ -778,17 +923,20 @@ sub generate_script {
 
     $file ||= catfile 't', 'TEST';
 
-    my $content = '';
+    my $body = "BEGIN { eval { require blib; } }\n";
+
+    $body .= Apache::TestConfig->modperl_2_inc_fixup;
+
     if (@Apache::TestMM::Argv) {
-        $content = "\%Apache::TestConfig::Argv = qw(@Apache::TestMM::Argv);\n";
+        $body .= "\n\%Apache::TestConfig::Argv = qw(@Apache::TestMM::Argv);\n";
     }
 
     my $header = Apache::TestConfig->perlscript_header;
 
-    $content .= join "\n",
-      $header, "use $class ();", "$class->new->run(\@ARGV);";
+    $body .= join "\n",
+        $header, "use $class ();", "$class->new->run(\@ARGV);";
 
-    Apache::Test::config()->write_perlscript($file, $content);
+    Apache::Test::config()->write_perlscript($file, $body);
 }
 
 # in idiomatic perl functions return 1 on success 0 on
@@ -806,3 +954,50 @@ sub exit_shell {
 }
 
 1;
+
+__END__
+
+=head1 NAME
+
+Apache::TestRun - Run the test suite
+
+=head1 SYNOPSIS
+
+
+=head1 DESCRIPTION
+
+The C<Apache::TestRun> package controls the configuration and running
+of the test suite.
+
+=head1 METHODS
+
+Several methods are sub-classable, if the default behavior should be
+changed.
+
+=head2 C<pre_configure>
+
+The C<pre_configure()> method is executed before the configuration for
+C<Apache::Test> is generated. So if you need to adjust the setup
+before I<httpd.conf> and other files are autogenerated, this is the
+right place to do so.
+
+For example if you don't want to inherit a LoadModule directive for
+I<mod_apreq.so> but to make sure that the local version is used, you
+can sub-class C<Apache::TestRun> and override this method in
+I<t/TEST.PL>:
+
+  package My::TestRun;
+  use base 'Apache::TestRun';
+  use Apache::TestConfig;
+  __PACKAGE__->new->run(@ARGV);
+  
+  sub pre_configure {
+      my $self = shift;
+      # Don't load an installed mod_apreq
+      Apache::TestConfig::autoconfig_skip_module_add('mod_apreq.c');
+  }
+
+Notice that the extension is I<.c>, and not I<.so>.
+
+
+=cut
