@@ -8,11 +8,9 @@ no warnings qw(redefine); # XXX, this should go away in production!
 
 our $VERSION = '1.99';
 
-use ModPerl::RegistryCooker;
-@ModPerl::RegistryBB::ISA = qw(ModPerl::RegistryCooker);
+use base qw(ModPerl::RegistryCooker);
 
-# META: prototyping ($$) segfaults on request
-sub handler {
+sub handler : method {
     my $class = (@_ >= 2) ? shift : __PACKAGE__;
     my $r = shift;
     return $class->new($r)->default_handler();
