@@ -156,8 +156,8 @@ my %defines_wanted = (
         satisfy    => [qw{SATISFY_}],
         remotehost => [qw{REMOTE_}],
         http       => [qw{HTTP_}],
-#       config     => [qw{DECLINE_CMD}],
-#       types      => [qw{DIR_MAGIC_TYPE}],
+        config     => [qw{DECLINE_CMD}],
+        types      => [qw{DIR_MAGIC_TYPE}],
         override   => [qw{OR_ ACCESS_CONF RSRC_CONF}],
         log        => [qw(APLOG_)],
     },
