@@ -5,7 +5,7 @@ use warnings FATAL => 'all';
 
 use mod_perl 1.99;
 
-our $VERSION = '0.08';
+our $VERSION = '0.09';
 
 use Apache::Const -compile => qw(OK);
 
@@ -63,6 +63,10 @@ sub handler {
 
     my $TouchFile = ref($o) && $o->dir_config("ReloadTouchFile");
 
+    my $ConstantRedefineWarnings = ref($o) && 
+        (lc($o->dir_config("ReloadConstantRedefineWarnings") || '') eq 'off') 
+            ? 0 : 1;
+
     my $TouchModules;
 
     if ($TouchFile) {
@@ -72,7 +76,7 @@ sub handler {
         $TouchTime = $touch_mtime;
         open my $fh, $TouchFile or die "Can't open '$TouchFile': $!";
         $TouchModules = <$fh>;
-        chomp $TouchModules;
+        chomp $TouchModules if $TouchModules;
     }
 
     if (ref($o) && (lc($o->dir_config("ReloadAll") || 'on') eq 'on')) {
@@ -141,6 +145,8 @@ sub handler {
                 undef %{$symref};
             }
             no warnings FATAL => 'all';
+            local $SIG{__WARN__} = \&skip_redefine_const_sub_warn
+                unless $ConstantRedefineWarnings;
             require $key;
             warn("Apache::Reload: process $$ reloading $key\n")
                     if $DEBUG;
@@ -151,5 +157,10 @@ sub handler {
     return Apache::OK;
 }
 
+sub skip_redefine_const_sub_warn {
+    return if $_[0] =~ /^Constant subroutine [\w:]+ redefined at/;
+    CORE::warn(@_);
+}
+
 1;
 __END__
