@@ -73,10 +73,14 @@ sub parse {
                 $val = File::Spec->canonpath(File::Spec->rel2abs($val));
             }
 
-            # MP_AP_PREFIX may not contain spaces
-            if ($key eq 'MP_AP_PREFIX' && Apache::Build::WIN32()) {
-                require Win32;
-                $val = Win32::GetShortPathName($val);
+            if ($key eq 'MP_AP_PREFIX') {
+                $val = File::Spec->canonpath(File::Spec->rel2abs($val));
+
+                if (Apache::Build::WIN32()) {
+                    # MP_AP_PREFIX may not contain spaces
+                    require Win32;
+                    $val = Win32::GetShortPathName($val);
+                }
             }
 
             if ($self->{$key}) {
