@@ -2,7 +2,10 @@ package ModPerl::Code;
 
 use strict;
 use warnings FATAL => 'all';
+
 use Config;
+use File::Spec::Functions qw(catfile catdir);
+
 use mod_perl ();
 use Apache::Build ();
 
@@ -34,27 +37,32 @@ my %hook_proto = (
     Process    => {
         ret  => 'void',
         args => [{type => 'apr_pool_t', name => 'p'},
-                 {type => 'server_rec', name => 's'}],
+                 {type => 'server_rec', name => 's'},
+                 {type => 'dummy', name => 'MP_HOOK_VOID'}],
     },
     Files      => {
         ret  => 'int',
         args => [{type => 'apr_pool_t', name => 'pconf'},
                  {type => 'apr_pool_t', name => 'plog'},
                  {type => 'apr_pool_t', name => 'ptemp'},
-                 {type => 'server_rec', name => 's'}],
+                 {type => 'server_rec', name => 's'},
+                 {type => 'dummy', name => 'MP_HOOK_RUN_ALL'}],
     },
     PerSrv     => {
         ret  => 'int',
-        args => [{type => 'request_rec', name => 'r'}],
+        args => [{type => 'request_rec', name => 'r'}, 
+                 {type => 'dummy', name => 'MP_HOOK_RUN_ALL'}],
     },
     Connection => {
         ret  => 'int',
-        args => [{type => 'conn_rec', name => 'c'}],
+        args => [{type => 'conn_rec', name => 'c'},
+                 {type => 'dummy', name => 'MP_HOOK_RUN_FIRST'}],
     },
     PreConnection => {
         ret  => 'int',
         args => [{type => 'conn_rec', name => 'c'},
-                 {type => 'void', name => 'csd'}],
+                 {type => 'void', name => 'csd'},
+                 {type => 'dummy', name => 'MP_HOOK_RUN_ALL'}],
     },
 );
 
@@ -108,7 +116,7 @@ my %flags = (
     Req => [qw(NONE SET_GLOBAL_REQUEST PARSE_HEADERS SETUP_ENV 
                CLEANUP_REGISTERED)],
     Interp => [qw(NONE IN_USE PUTBACK CLONED BASE)],
-    Handler => [qw(NONE PARSED METHOD OBJECT ANON AUTOLOAD DYNAMIC)],
+    Handler => [qw(NONE PARSED METHOD OBJECT ANON AUTOLOAD DYNAMIC FAKE)],
 );
 
 $flags{DirSeen} = $flags{Dir};
@@ -208,6 +216,12 @@ sub generate_handler_hooks {
             my($protostr, $pass) = canon_proto($prototype, $name);
             my $ix = $self->{handler_index}->{$class}->[$i];
 
+            if ($callback =~ m/modperl_callback_per_(dir|srv)/) {
+                if ($ix =~ m/AUTH|TYPE|TRANS/) {
+                    $pass =~ s/MP_HOOK_RUN_ALL/MP_HOOK_RUN_FIRST/;
+                }
+            }
+
             print $h_fh "\n$protostr;\n";
 
             print $c_fh <<EOF;
@@ -440,13 +454,15 @@ my %trace = (
 #    'a' => 'all',
     'c' => 'configuration for directive handlers',
     'd' => 'directive processing',
-    's' => 'perl sections',
-    'h' => 'handlers',
-    'm' => 'memory allocations',
-    't' => 'benchmark-ish timings',
-    'i' => 'interpreter pool management',
-    'g' => 'Perl runtime interaction',
+    'e' => 'environment variables',
     'f' => 'filters',
+    'g' => 'Perl runtime interaction',
+    'h' => 'handlers',
+    'i' => 'interpreter pool management',
+    'm' => 'memory allocations',
+    'o' => 'I/O',
+    's' => 'perl sections',
+    't' => 'benchmark-ish timings',
 );
 
 sub generate_trace {
@@ -554,8 +570,12 @@ sub canon_define {
 
 sub canon_args {
     my $args = shift->{args};
-    my @in   = map { "$_->{type} *$_->{name}" } @$args;
     my @pass = map { $_->{name} } @$args;
+    my @in;
+    foreach my $href (@$args) {
+        push @in, "$href->{type} *$href->{name}"
+            unless $href->{type} eq 'dummy';
+    }
     return wantarray ? (\@in, \@pass) : \@in;
 }
 
@@ -587,8 +607,8 @@ my %sources = (
 );
 
 my @c_src_names = qw(interp tipool log config cmd options callback handler
-                     gtop util io filter bucket mgv pcw global env cgi
-                     perl perl_global perl_pp sys module svptr_table
+                     gtop util io io_apache filter bucket mgv pcw global env
+                     cgi perl perl_global perl_pp sys module svptr_table
                      const constants apache_compat);
 my @h_src_names = qw(perl_unembed);
 my @g_c_names = map { "modperl_$_" } qw(hooks directives flags xsinit);
@@ -719,6 +739,8 @@ sub generate {
     #create bootstrap method for static xs modules
     my $static_xs = [keys %{ $build->{XS} }];
     ExtUtils::Embed::xsinit($xsinit, 1, $static_xs);
+
+    #$self->generate_constants_pod();
 }
 
 my $constant_prefixes = join '|', qw{APR?};
@@ -769,7 +791,7 @@ sub constants_lookup_code {
     my $package_len = length $package;
 
     my $func = canon_func(qw(constants lookup), $postfix);
-    my $proto = "int $func(const char *name)";
+    my $proto = "SV \*$func(pTHX_ const char *name)";
 
     print $h_fh "$proto;\n";
 
@@ -804,7 +826,20 @@ EOF
             print $c_fh <<EOF;
 $ifdef[0]
           if (strEQ(name, "$name")) {
-              return $alias{$name};
+EOF
+
+            if ($name eq 'DECLINE_CMD' || $name eq 'DIR_MAGIC_TYPE') {
+                print $c_fh <<EOF;
+              return newSVpv($alias{$name}, 0);
+EOF
+            }
+            else {
+                print $c_fh <<EOF;
+              return newSViv($alias{$name});
+EOF
+            }
+
+            print $c_fh <<EOF;
           }
 $ifdef[1]
 EOF
@@ -814,8 +849,8 @@ EOF
 
     print $c_fh <<EOF
     };
-    Perl_croak_nocontext("unknown constant %s", name);
-    return MP_ENOCONST;
+    Perl_croak(aTHX_ "unknown $class\:: constant %s", name);
+    return newSViv(MP_ENOCONST);
 }
 EOF
 }
@@ -880,12 +915,135 @@ sub constants_group_lookup_code {
 
     print $c_fh <<EOF;
     };
-    Perl_croak_nocontext("unknown group `%s'", name);
+    Perl_croak_nocontext("unknown $class\:: group `%s'", name);
     return NULL;
 }
 EOF
 }
 
+my %seen_const = ();
+# generates APR::Const and Apache::Const manpages in ./tmp/
+sub generate_constants_pod {
+    my($self) = @_;
+
+    my %data = ();
+    generate_constants_group_lookup_doc(\%data);
+    generate_constants_lookup_doc(\%data);
+
+    # XXX: may be dump %data into ModPerl::MethodLookup and provide an
+    # easy api to map const groups to constants and vice versa
+
+    require File::Path;
+    my $file = "Const.pod";
+    for my $class (keys %data) {
+        my $path = catdir "tmp", $class;
+        File::Path::mkpath($path, 0, 0755);
+        my $filepath = catfile $path, $file;
+        open my $fh, ">$filepath" or die "Can't open $filepath: $!\n";
+
+        print $fh <<"EOF";
+=head1 NAME
+
+$class\::Const - Perl Interface for $class Constants
+
+=head1 SYNOPSIS
+
+=head1 CONSTANTS
+
+EOF
+
+        my $groups = $data{$class};
+        for my $group (sort keys %$groups) {
+            print $fh <<"EOF";
+
+
+
+=head2 C<:$group>
+
+  use $class\::Const -compile qw(:$group);
+
+The C<:$group> group is for XXX constants.
+
+EOF
+
+            for my $const (sort @{ $groups->{$group} }) {
+                print $fh "=head3 C<$class\::$const>\n\n\n";
+            }
+        }
+
+        print $fh "=cut\n";
+    }
+}
+
+sub generate_constants_lookup_doc {
+    my($data) = @_;
+
+    while (my($class, $groups) = each %$Apache::ConstantsTable) {
+        my $constants = [map { @$_ } values %$groups];
+
+        constants_lookup_code_doc($constants, $class, $data);
+    }
+}
+
+sub generate_constants_group_lookup_doc {
+    my($data) = @_;
+
+    while (my($class, $groups) = each %$Apache::ConstantsTable) {
+        constants_group_lookup_code_doc($class, $groups, $data);
+    }
+}
+
+sub constants_group_lookup_code_doc {
+    my($class, $groups, $data) = @_;
+    my @tags;
+    my @code;
+
+    while (my($group, $constants) = each %$groups) {
+        $data->{$class}{$group} = [
+            map {
+                my @ifdef = constants_ifdef($_);
+                s/^($constant_prefixes)_?//o;
+                $seen_const{$class}{$_}++;
+                $_;
+            } @$constants
+        ];
+    }
+}
+
+sub constants_lookup_code_doc {
+    my($constants, $class, $data) = @_;
+
+    my(%switch, %alias);
+
+    %alias = %shortcuts;
+
+    my $postfix = lc $class;
+    my $package = $class . '::';
+    my $package_len = length $package;
+
+    my $func = canon_func(qw(constants lookup), $postfix);
+
+    for (@$constants) {
+        if (s/^($constant_prefixes)(_)?//o) {
+            $alias{$_} = join $2 || "", $1, $_;
+        }
+        else {
+            $alias{$_} ||= $_;
+        }
+        next unless /^([A-Z])/;
+        push @{ $switch{$1} }, $_;
+    }
+
+    for my $key (sort keys %switch) {
+        my $names = $switch{$key};
+        for my $name (@$names) {
+            my @ifdef = constants_ifdef($alias{$name});
+            push @{ $data->{$class}{other} }, $name
+                unless $seen_const{$class}{$name}
+        }
+    }
+}
+
 1;
 __END__
 
