@@ -9,6 +9,71 @@ static int MP_init_status = 0;
 #define MP_IS_STARTING    (MP_init_status == 1 ? 1 : 0)
 #define MP_IS_RUNNING     (MP_init_status == 2 ? 1 : 0)
 
+#if !(PERL_REVISION == 5 && ( PERL_VERSION < 8 || \
+    (PERL_VERSION == 8 && PERL_SUBVERSION == 0))) && \
+    (defined(USE_HASH_SEED) || defined(USE_HASH_SEED_EXPLICIT))
+#define MP_NEED_HASH_SEED_FIXUP
+#endif
+
+
+#ifdef MP_NEED_HASH_SEED_FIXUP
+static UV   MP_init_hash_seed = 0;
+static bool MP_init_hash_seed_set = FALSE;
+#endif
+
+/* see modperl_hash_seed_set() */
+static void modperl_hash_seed_init(apr_pool_t *p) 
+{
+#ifdef MP_NEED_HASH_SEED_FIXUP
+    char *s;
+    /* check if there is a specific hash seed passed via the env var
+     * and if that's the case -- use it */
+    apr_status_t rv = apr_env_get(&s, "PERL_HASH_SEED", p);
+    if (rv == APR_SUCCESS) {
+        if (s) {
+            while (isSPACE(*s)) s++;
+        }
+        if (s && isDIGIT(*s)) {
+            MP_init_hash_seed = (UV)Atol(s); // XXX: Atoul()?
+            MP_init_hash_seed_set = TRUE;
+        }
+    }
+    
+    /* calculate our own random hash seed */
+    if (!MP_init_hash_seed_set) {
+        apr_uuid_t *uuid = (apr_uuid_t *)apr_palloc(p, sizeof(apr_uuid_t));
+        char buf[APR_UUID_FORMATTED_LENGTH + 1];
+        int i;
+
+        apr_initialize();
+        apr_uuid_get(uuid);
+        apr_uuid_format(buf, uuid);
+        /* fprintf(stderr, "UUID: %s\n", buf); */
+
+        /* XXX: need a better alg to convert uuid string into a seed */
+        for (i=0; buf[i]; i++){
+            MP_init_hash_seed += (UV)(i+1)*(buf[i]+MP_init_hash_seed);
+        }
+
+        MP_init_hash_seed_set = TRUE;
+    }
+#endif
+}
+
+/* before 5.8.1, perl was using HASH_SEED=0, starting from 5.8.1
+ * it randomizes if perl was compiled with ccflags -DUSE_HASH_SEED
+ * or -DUSE_HASH_SEED_EXPLICIT, in which case we need to tell perl
+ * to use the same seed everywhere */
+static void modperl_hash_seed_set(pTHX) 
+{
+#ifdef MP_NEED_HASH_SEED_FIXUP
+    if (MP_init_hash_seed_set) {
+        PL_hash_seed_set = TRUE;
+        PL_hash_seed = MP_init_hash_seed;
+    }
+#endif
+}
+
 #ifndef USE_ITHREADS
 static apr_status_t modperl_shutdown(void *data)
 {
@@ -147,8 +212,7 @@ PerlInterpreter *modperl_startup(server_rec *s, apr_pool_t *p)
 
 #ifdef MP_USE_GTOP
     MP_TRACE_m_do(
-        scfg->gtop = modperl_gtop_new(p);
-        modperl_gtop_do_proc_mem_before(MP_FUNC ": perl_parse");
+        modperl_gtop_do_proc_mem_before(MP_FUNC, "perl_parse");
     );
 #endif
 
@@ -165,6 +229,10 @@ PerlInterpreter *modperl_startup(server_rec *s, apr_pool_t *p)
 
     perl_construct(perl);
 
+    modperl_hash_seed_set(aTHX);
+
+    modperl_io_apache_init(aTHX);
+
     PL_perl_destruct_level = 2;
 
     MP_boot_data_set(p, s);
@@ -203,7 +271,7 @@ PerlInterpreter *modperl_startup(server_rec *s, apr_pool_t *p)
 
 #ifdef MP_USE_GTOP
     MP_TRACE_m_do(
-        modperl_gtop_do_proc_mem_after(MP_FUNC ": perl_parse");
+        modperl_gtop_do_proc_mem_after(MP_FUNC, "perl_parse");
     );
 #endif
 
@@ -297,10 +365,11 @@ int modperl_init_vhost(server_rec *s, apr_pool_t *p,
         }   
 #endif
         
-        if (!modperl_config_apply_PerlModule(s, scfg, perl, p)) {
+        if (!modperl_config_apply_PerlRequire(s, scfg, perl, p)) {
             return HTTP_INTERNAL_SERVER_ERROR;
         }
-        if (!modperl_config_apply_PerlRequire(s, scfg, perl, p)) {
+
+        if (!modperl_config_apply_PerlModule(s, scfg, perl, p)) {
             return HTTP_INTERNAL_SERVER_ERROR;
         }
     }
@@ -507,9 +576,10 @@ int modperl_is_running(void)
 int modperl_hook_pre_config(apr_pool_t *p, apr_pool_t *plog,
                             apr_pool_t *ptemp)
 {
-    /* XXX: htf can we have PerlPreConfigHandler
-     * without first configuring mod_perl ?
-     */
+    /* we can't have PerlPreConfigHandler without first configuring mod_perl */
+
+    /* perl 5.8.1+ */
+    modperl_hash_seed_init(p);
 
     return OK;
 }
@@ -528,16 +598,31 @@ static int modperl_hook_post_config(apr_pool_t *pconf, apr_pool_t *plog,
     MP_dSCFG(s);
     dTHXa(scfg->mip->parent->perl);
 #endif
+
+#ifdef MP_TRACE
+    /* httpd core open_logs handler re-opens s->error_log, which might
+     * change, even though it still points to the same physical file
+     * (.e.g on win32 the filehandle will be different. Therefore
+     * reset the tracing logfile setting here, since this is the
+     * earliest place, happening after the open_logs phase */
+    modperl_trace_logfile_set(s->error_log);
+#endif
+    
     ap_add_version_component(pconf, MP_VERSION_STRING);
     ap_add_version_component(pconf,
                              Perl_form(aTHX_ "Perl/v%vd", PL_patchlevel));
     modperl_mgv_hash_handlers(pconf, s);
-    modperl_modglobal_hash_keys();
-    modperl_env_hash_keys();
+    modperl_modglobal_hash_keys(aTHX);
+    modperl_env_hash_keys(aTHX);
 #ifdef USE_ITHREADS
     modperl_init_clones(s, pconf);
 #endif
 
+#ifdef MP_NEED_HASH_SEED_FIXUP
+    ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
+                 "mod_perl: using Perl HASH_SEED: %"UVuf, MP_init_hash_seed);
+#endif
+    
     return OK;
 }
 
@@ -585,7 +670,7 @@ static apr_status_t modperl_child_exit(void *data)
     char *level = NULL;
     server_rec *s = (server_rec *)data;
     
-    modperl_callback_process(MP_CHILD_EXIT_HANDLER, server_pool, s);
+    modperl_callback_process(MP_CHILD_EXIT_HANDLER, server_pool, s, MP_HOOK_VOID);
     
     if ((level = getenv("PERL_DESTRUCT_LEVEL"))) {
         modperl_destruct_level = atoi(level);
@@ -612,12 +697,7 @@ static void modperl_hook_child_init(apr_pool_t *p, server_rec *s)
                               apr_pool_cleanup_null);
 }
 
-/* api change in 2.0.40-ish */
-#if (MODULE_MAGIC_NUMBER_MAJOR >= 20020628)
-#   define MP_FILTER_HANDLER(f) f, NULL
-#else
-#   define MP_FILTER_HANDLER(f) f
-#endif
+#define MP_FILTER_HANDLER(f) f, NULL
 
 void modperl_register_hooks(apr_pool_t *p)
 {
@@ -774,7 +854,7 @@ static int modperl_response_handler_run(request_rec *r, int finish)
 
     modperl_response_init(r);
 
-    retval = modperl_callback_per_dir(MP_RESPONSE_HANDLER, r);
+    retval = modperl_callback_per_dir(MP_RESPONSE_HANDLER, r, MP_HOOK_RUN_FIRST);
 
     if ((retval == DECLINED) && r->content_type) {
         r->handler = r->content_type; /* let http_core or whatever try */
@@ -834,20 +914,20 @@ int modperl_response_handler_cgi(request_rec *r)
         modperl_global_request_set(r);
     }
 
-    h_stdout = modperl_io_tie_stdout(aTHX_ r);
-    h_stdin  = modperl_io_tie_stdin(aTHX_ r);
+    h_stdin  = modperl_io_override_stdin(aTHX_ r);
+    h_stdout = modperl_io_override_stdout(aTHX_ r);
 
     modperl_env_request_tie(aTHX_ r);
 
     retval = modperl_response_handler_run(r, FALSE);
 
-    modperl_io_handle_untie(aTHX_ h_stdout);
-    modperl_io_handle_untie(aTHX_ h_stdin);
-
     modperl_env_request_untie(aTHX_ r);
 
     modperl_perl_global_request_restore(aTHX_ r);
 
+    modperl_io_restore_stdin(aTHX_ h_stdin);
+    modperl_io_restore_stdout(aTHX_ h_stdout);
+
 #ifdef USE_ITHREADS
     if (MpInterpPUTBACK(interp)) {
         /* PerlInterpScope handler */
