@@ -20,13 +20,7 @@
 #include "ap_mpm.h"
 
 #include "apr_version.h"
-#ifndef APR_POLLIN
-/*
- * apr_poll.h introduced around 2.0.40
- * APR_POLL* constants moved here around 2.0.44
- */
 #include "apr_poll.h"
-#endif
 #include "apr_lib.h"
 #include "apr_strings.h"
 #include "apr_uri.h"
@@ -34,7 +28,9 @@
 #include "apr_buckets.h"
 #include "apr_time.h"
 #include "apr_network_io.h"
-
+#include "apr_general.h"
+#include "apr_uuid.h"
+#include "apr_env.h"
 #include "util_filter.h"
 
 #include "util_script.h"
