@@ -6,7 +6,7 @@ int modperl_callback(pTHX_ modperl_handler_t *handler, apr_pool_t *p,
     CV *cv=Nullcv;
     I32 flags = G_EVAL|G_SCALAR;
     dSP;
-    int count, status;
+    int count, status = OK;
 
     if ((status = modperl_handler_resolve(aTHX_ &handler, p, s)) != OK) {
         return status;
@@ -45,28 +45,46 @@ int modperl_callback(pTHX_ modperl_handler_t *handler, apr_pool_t *p,
             cv = modperl_mgv_cv(gv);
         }
         else {
-            char *name = modperl_mgv_as_string(aTHX_ handler->mgv_cv, p, 0);
-            MP_TRACE_h(MP_FUNC, "lookup of %s failed\n", name);
+            
+            const char *name;
+            modperl_mgv_t *symbol = handler->mgv_cv;
+            
+             /* XXX: need to validate *symbol */
+            if (symbol && symbol->name) {
+                name = modperl_mgv_as_string(aTHX_ symbol, p, 0);
+            }
+            else {
+                name = handler->name;
+            }
+            
+            MP_TRACE_h(MP_FUNC, "[%s %s] lookup of %s failed\n",
+                         modperl_pid_tid(p), modperl_server_desc(s, p), name);
+            ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
+                         "lookup of '%s' failed\n", name);
+            status = HTTP_INTERNAL_SERVER_ERROR;
         }
     }
 
-    count = call_sv((SV*)cv, flags);
+    if (status == OK) {
+        count = call_sv((SV*)cv, flags);
 
-    SPAGAIN;
+        SPAGAIN;
 
-    if (count != 1) {
-        status = OK;
-    }
-    else {
-        status = POPi;
-        /* assume OK for non-http status codes and for 200 (HTTP_OK) */
-        if (((status > 0) && (status < 100)) ||
-            (status == 200) || (status > 600)) {
+        if (count != 1) {
             status = OK;
         }
-    }
+        else {
+            status = POPi;
+            /* assume OK for non-http status codes and for 200 (HTTP_OK) */
+            if (((status > 0) && (status < 100)) ||
+                (status == 200) || (status > 600)) {
+                status = OK;
+            }
+        }
 
-    PUTBACK;
+        PUTBACK;
+    }
+    
     FREETMPS;LEAVE;
 
     if (SvTRUE(ERRSV)) {
@@ -81,7 +99,8 @@ int modperl_callback_run_handlers(int idx, int type,
                                   request_rec *r, conn_rec *c, server_rec *s,
                                   apr_pool_t *pconf,
                                   apr_pool_t *plog,
-                                  apr_pool_t *ptemp)
+                                  apr_pool_t *ptemp,
+                                  modperl_hook_run_mode_e run_mode)
 {
 #ifdef USE_ITHREADS
     pTHX;
@@ -180,7 +199,8 @@ int modperl_callback_run_handlers(int idx, int type,
      * different handler e.g. jumping from 'modperl' to 'perl-script',
      * before calling push_handler */
     nelts = av->nelts;
-    MP_TRACE_h(MP_FUNC, "running %d %s handlers\n", nelts, desc);
+    MP_TRACE_h(MP_FUNC, "[%s] running %d %s handlers\n",
+               modperl_pid_tid(p), nelts, desc);
     handlers = (modperl_handler_t **)av->elts;
 
     for (i=0; i<nelts; i++) {
@@ -188,71 +208,117 @@ int modperl_callback_run_handlers(int idx, int type,
         
         MP_TRACE_h(MP_FUNC, "%s returned %d\n", handlers[i]->name, status);
 
-        if ((status != OK) && (status != DECLINED)) {
-            status = modperl_errsv(aTHX_ status, r, s);
-#ifdef MP_TRACE
-            if (i+1 != nelts) {
-                MP_TRACE_h(MP_FUNC, "there were %d uncalled handlers\n",
-                           nelts-i-1);
-            }
-#endif
-            break;
-        }
+        /* follow Apache's lead and let OK terminate the phase for
+         * MP_HOOK_RUN_FIRST handlers.  MP_HOOK_RUN_ALL handlers keep going on OK.
+         * MP_HOOK_VOID handler ignore all errors.
+         */
 
+        if (run_mode == MP_HOOK_RUN_ALL) {
+            /* the normal case:
+             *   OK and DECLINED continue 
+             *   errors end the phase
+             */
+            if ((status != OK) && (status != DECLINED)) {
+
+                status = modperl_errsv(aTHX_ status, r, s);
+#ifdef MP_TRACE
+                if (i+1 != nelts) {
+                    MP_TRACE_h(MP_FUNC, "error status %d leaves %d uncalled handlers\n",
+                               status, desc, nelts-i-1);
+                }
+#endif
+                break;
+            }
+        }
+        else if (run_mode == MP_HOOK_RUN_FIRST) {
+            /* the exceptional case:
+             *   OK and errors end the phase
+             *   DECLINED continues
+             */
+
+            if (status == OK) {
+#ifdef MP_TRACE
+                if (i+1 != nelts) {
+                    MP_TRACE_h(MP_FUNC, "OK ends the %s stack, leaving %d uncalled handlers\n",
+                               desc, nelts-i-1);
+                }
+#endif
+                break;
+            }
+            if (status != DECLINED) {
+                status = modperl_errsv(aTHX_ status, r, s);
+#ifdef MP_TRACE
+                if (i+1 != nelts) {
+                    MP_TRACE_h(MP_FUNC, "error status %d leaves %d uncalled handlers\n",
+                               status, desc, nelts-i-1);
+                }
+#endif
+                break;
+            }
+        }
+        else {
+            /* the rare case.
+             * MP_HOOK_VOID handlers completely ignore the return status
+             * Apache should handle whatever mod_perl returns, 
+             * so there is no need to mess with the status
+             */
+        }
     }
 
     SvREFCNT_dec((SV*)av_args);
 
-#ifdef USE_ITHREADS
-    if (interp && MpInterpPUTBACK(interp)) {
-        /* PerlInterpScope handler */
-        modperl_interp_unselect(interp);
-    }
-#endif
+    /* PerlInterpScope handler */
+    MP_INTERP_PUTBACK(interp);
 
     return status;
 }
 
-int modperl_callback_per_dir(int idx, request_rec *r)
+int modperl_callback_per_dir(int idx, request_rec *r,
+                             modperl_hook_run_mode_e run_mode)
 {
     return modperl_callback_run_handlers(idx, MP_HANDLER_TYPE_PER_DIR,
                                          r, NULL, r->server,
-                                         NULL, NULL, NULL);
+                                         NULL, NULL, NULL, run_mode);
 }
 
-int modperl_callback_per_srv(int idx, request_rec *r)
+int modperl_callback_per_srv(int idx, request_rec *r, 
+                             modperl_hook_run_mode_e run_mode)
 {
     return modperl_callback_run_handlers(idx, MP_HANDLER_TYPE_PER_SRV,
                                          r, NULL, r->server,
-                                         NULL, NULL, NULL);
+                                         NULL, NULL, NULL, run_mode);
 }
 
-int modperl_callback_connection(int idx, conn_rec *c)
+int modperl_callback_connection(int idx, conn_rec *c, 
+                                modperl_hook_run_mode_e run_mode)
 {
     return modperl_callback_run_handlers(idx, MP_HANDLER_TYPE_CONNECTION,
                                          NULL, c, c->base_server,
-                                         NULL, NULL, NULL);
+                                         NULL, NULL, NULL, run_mode);
 }
 
-int modperl_callback_pre_connection(int idx, conn_rec *c, void *csd)
+int modperl_callback_pre_connection(int idx, conn_rec *c, void *csd,
+                                    modperl_hook_run_mode_e run_mode)
 {
     return modperl_callback_run_handlers(idx, MP_HANDLER_TYPE_PRE_CONNECTION,
                                          NULL, c, c->base_server,
-                                         NULL, NULL, NULL);
+                                         NULL, NULL, NULL, run_mode);
 }
 
-void modperl_callback_process(int idx, apr_pool_t *p, server_rec *s)
+void modperl_callback_process(int idx, apr_pool_t *p, server_rec *s,
+                              modperl_hook_run_mode_e run_mode)
 {
     modperl_callback_run_handlers(idx, MP_HANDLER_TYPE_PROCESS,
                                   NULL, NULL, s,
-                                  p, NULL, NULL);
+                                  p, NULL, NULL, run_mode);
 }
 
 int modperl_callback_files(int idx,
                            apr_pool_t *pconf, apr_pool_t *plog,
-                           apr_pool_t *ptemp, server_rec *s)
+                           apr_pool_t *ptemp, server_rec *s,
+                           modperl_hook_run_mode_e run_mode)
 {
     return modperl_callback_run_handlers(idx, MP_HANDLER_TYPE_FILES,
                                          NULL, NULL, s,
-                                         pconf, plog, ptemp);
+                                         pconf, plog, ptemp, run_mode);
 }
