@@ -112,6 +112,10 @@ modperl_config_srv_t *modperl_config_srv_new(apr_pool_t *p)
     scfg->PassEnv = apr_table_make(p, 2);
     scfg->SetEnv = apr_table_make(p, 2);
     
+#ifdef MP_USE_GTOP
+    scfg->gtop = modperl_gtop_new(p);
+#endif        
+    
     modperl_config_srv_argv_push((char *)ap_server_argv0);
 
     MP_TRACE_d(MP_FUNC, "0x%lx\n", (unsigned long)scfg);
@@ -169,7 +173,7 @@ void *modperl_config_srv_create(apr_pool_t *p, server_rec *s)
         /* Must store the global server record as early as possible,
          * because if mod_perl happens to be started from within a
          * vhost (e.g., PerlLoadModule) the base server record won't
-         * be availalbe to vhost and things will blow up
+         * be available to vhost and things will blow up
          */
         modperl_init_globals(s, p);
 
@@ -286,8 +290,14 @@ apr_status_t modperl_config_request_cleanup(pTHX_ request_rec *r)
         rcfg->pnotes = Nullhv;
     }
 
-    retval = modperl_callback_per_dir(MP_CLEANUP_HANDLER, r);
-    
+    retval = modperl_callback_per_dir(MP_CLEANUP_HANDLER, r, MP_HOOK_RUN_ALL);
+
+    /* undo changes to %ENV caused by +SetupEnv, perl-script, or
+     * $r->subprocess_env, so the values won't persist  */
+    if (MpReqSETUP_ENV(rcfg)) {
+        modperl_env_request_unpopulate(aTHX_ r);
+    }
+
     return retval;
 }
 
