@@ -5,11 +5,15 @@
 #define MP_FILTER_NAME_FORMAT "   %s\n\n\t"
 
 #define MP_FILTER_NAME(f) \
-    ((modperl_filter_ctx_t *)f->ctx)->handler->name
+    (is_modperl_filter(f) \
+        ? ((modperl_filter_ctx_t *)(f)->ctx)->handler->name \
+        : (f)->frec->name)
 
 #define MP_FILTER_TYPE(filter) \
-    ((modperl_filter_ctx_t *)filter->f->ctx)->handler->attrs & \
-        MP_FILTER_CONNECTION_HANDLER  ? "connection" : "request"
+    (is_modperl_filter(filter->f) \
+        ? ((modperl_filter_ctx_t *)(filter)->f->ctx)->handler->attrs & \
+            MP_FILTER_CONNECTION_HANDLER  ? "connection" : "request" \
+        : "unknown")
 
 #define MP_FILTER_MODE(filter) \
     (filter->mode == MP_INPUT_FILTER_MODE ? "input" : "output")
@@ -17,7 +21,7 @@
 #define MP_FILTER_POOL(f) f->r ? f->r->pool : f->c->pool
 
 /* this function is for tracing only, it's not optimized for performance */
-static const char* next_filter_name(ap_filter_t *f)
+static int is_modperl_filter(ap_filter_t *f)
 {
     const char *name = f->frec->name;
 
@@ -26,19 +30,22 @@ static const char* next_filter_name(ap_filter_t *f)
         !strcasecmp(name, MP_FILTER_CONNECTION_OUTPUT_NAME) ||
         !strcasecmp(name, MP_FILTER_REQUEST_INPUT_NAME)     ||
         !strcasecmp(name, MP_FILTER_REQUEST_OUTPUT_NAME) ) {
-        return ((modperl_filter_ctx_t *)f->ctx)->handler->name;
+        return 1;
     }
     else {
-        return name;
+        return 0;
     }
 }
 
+
 MP_INLINE static apr_status_t send_input_eos(modperl_filter_t *filter)
 {
     apr_bucket_alloc_t *ba = filter->f->c->bucket_alloc;
     apr_bucket *b = apr_bucket_eos_create(ba);
     APR_BRIGADE_INSERT_TAIL(filter->bb_out, b);
     ((modperl_filter_ctx_t *)filter->f->ctx)->sent_eos = 1;
+    MP_TRACE_f(MP_FUNC, MP_FILTER_NAME_FORMAT
+               "write out: EOS bucket\n", MP_FILTER_NAME(filter->f));
     return APR_SUCCESS;
 }
 
@@ -47,6 +54,8 @@ MP_INLINE static apr_status_t send_input_flush(modperl_filter_t *filter)
     apr_bucket_alloc_t *ba = filter->f->c->bucket_alloc;
     apr_bucket *b = apr_bucket_flush_create(ba);
     APR_BRIGADE_INSERT_TAIL(filter->bb_out, b);
+    MP_TRACE_f(MP_FUNC, MP_FILTER_NAME_FORMAT
+               "write out: FLUSH bucket\n", MP_FILTER_NAME(filter->f));
     return APR_SUCCESS;
 }
 
@@ -106,6 +115,10 @@ MP_INLINE apr_status_t modperl_wbucket_pass(modperl_wbucket_t *wb,
             work_buf = (char *)apr_pcalloc(wb->pool, sizeof(char*)*len);
             memcpy((void*)work_buf, buf, len);
         }
+
+        MP_TRACE_f(MP_FUNC, "\n\n\tparsing headers: %d bytes [%s]\n", len,
+                   apr_pstrmemdup(wb->pool, work_buf, len));
+        
         status = modperl_cgi_header_parse(r, (char *)work_buf, &bodytext);
 
         wb->header_parse = 0; /* only once per-request */
@@ -117,8 +130,9 @@ MP_INLINE apr_status_t modperl_wbucket_pass(modperl_wbucket_t *wb,
             ap_log_error(APLOG_MARK, APLOG_WARNING,
                          0, r->server, "%s did not send an HTTP header",
                          r->uri);
+            r->status = status;
             /* XXX: bodytext == NULL here */
-            return status;
+            return APR_SUCCESS;
         }
         else if (!bodytext) {
             return APR_SUCCESS;
@@ -143,18 +157,21 @@ MP_INLINE apr_status_t modperl_wbucket_pass(modperl_wbucket_t *wb,
     MP_TRACE_f(MP_FUNC, "\n\n\twrite out: %d bytes\n"
                "\t\tfrom %s\n\t\tto %s filter handler\n",
                len, 
-               (wb->r && wb->filters == &wb->r->output_filters)
-                   ? "response handler" : "current filter handler",
-               next_filter_name(*(wb->filters)));
+               ((wb->r && wb->filters == &wb->r->output_filters)
+                   ? "response handler" : "current filter handler"),
+               MP_FILTER_NAME(*(wb->filters)));
 
     return ap_pass_brigade(*(wb->filters), bb);
 }
 
-/* if add_flush_bucket is TRUE
- *  and there is data to flush,
- *       a flush bucket is added to the tail of bb with data
- * otherwise
- *       a flush bucket is sent in its own bb
+/* flush data if any,
+ * if add_flush_bucket is TRUE
+ *     if there is data to flush
+ *         a flush bucket is added to the tail of bb with data
+ *     else
+ *         a flush bucket is sent in its own bb
+ * else
+ *     nothing is sent
  */
 MP_INLINE apr_status_t modperl_wbucket_flush(modperl_wbucket_t *wb,
                                              int add_flush_bucket)
@@ -169,7 +186,7 @@ MP_INLINE apr_status_t modperl_wbucket_flush(modperl_wbucket_t *wb,
     else if (add_flush_bucket) {
         rv = send_output_flush(*(wb->filters));
     }
-    
+
     return rv;
 }
 
@@ -180,8 +197,6 @@ MP_INLINE apr_status_t modperl_wbucket_write(pTHX_ modperl_wbucket_t *wb,
     apr_size_t len = *wlen;
     *wlen = 0;
 
-    MP_TRACE_f(MP_FUNC, "\n\n\tbuffer out: %d bytes\n", len);
-
     if ((len + wb->outcnt) > sizeof(wb->outbuf)) {
         apr_status_t rv;
         if ((rv = modperl_wbucket_flush(wb, FALSE)) != APR_SUCCESS) {
@@ -235,7 +250,8 @@ modperl_filter_t *modperl_filter_new(ap_filter_t *f,
     MP_TRACE_f(MP_FUNC, MP_FILTER_NAME_FORMAT
                "new: %s %s filter (0x%lx)\n",
                MP_FILTER_NAME(f),
-               MP_FILTER_TYPE(filter), MP_FILTER_MODE(filter),
+               MP_FILTER_TYPE(filter),
+               MP_FILTER_MODE(filter),
                (unsigned long)filter);
 
     return filter;
@@ -315,6 +331,7 @@ int modperl_filter_resolve_init_handler(pTHX_ modperl_handler_t *handler,
 }
 
 static int modperl_run_filter_init(ap_filter_t *f,
+                                   modperl_filter_mode_e mode,
                                    modperl_handler_t *handler) 
 {
     AV *args = Nullav;
@@ -333,15 +350,18 @@ static int modperl_run_filter_init(ap_filter_t *f,
                               "Apache::Filter", f,
                               NULL);
 
-    /* XXX: do we need it? */
-    /* modperl_filter_mg_set(aTHX_ AvARRAY(args)[0], filter); */
+    modperl_filter_mg_set(aTHX_ AvARRAY(args)[0],
+                          modperl_filter_new(f, NULL, mode, 0, 0, 0));
 
+    /* XXX filters are VOID handlers.  should we ignore the status? */
     if ((status = modperl_callback(aTHX_ handler, p, r, s, args)) != OK) {
         status = modperl_errsv(aTHX_ status, r, s);
     }
 
     SvREFCNT_dec((SV*)args);
 
+    MP_INTERP_PUTBACK(interp);
+
     MP_TRACE_f(MP_FUNC, MP_FILTER_NAME_FORMAT
                "return: %d\n", handler->name, status);
     
@@ -379,6 +399,7 @@ int modperl_run_filter(modperl_filter_t *filter)
         av_push(args, newSViv(filter->readbytes));
     }
 
+    /* XXX filters are VOID handlers.  should we ignore the status? */
     if ((status = modperl_callback(aTHX_ handler, p, r, s, args)) != OK) {
         status = modperl_errsv(aTHX_ status, r, s);
     }
@@ -411,6 +432,8 @@ int modperl_run_filter(modperl_filter_t *filter)
         MP_FAILURE_CROAK(modperl_output_filter_flush(filter));
     }
 
+    MP_INTERP_PUTBACK(interp);
+
     MP_TRACE_f(MP_FUNC, MP_FILTER_NAME_FORMAT
                "return: %d\n", handler->name, status);
     
@@ -652,16 +675,15 @@ MP_INLINE apr_status_t modperl_input_filter_flush(modperl_filter_t *filter)
         /* no data should be sent after EOS has been sent */
         return filter->rc;
     }
-    
-    if (filter->eos || filter->flush) {
-        MP_TRACE_f(MP_FUNC, MP_FILTER_NAME_FORMAT
-                   "write out: %s bucket\n",
-                   MP_FILTER_NAME(filter->f),
-                   filter->eos ? "EOS" : "FLUSH");
-        filter->rc = filter->eos ?
-            send_input_eos(filter) : send_input_flush(filter);
-        /* modperl_brigade_dump(filter->bb_out, stderr); */
-        filter->flush = filter->eos = 0;
+
+    if (filter->flush) {
+        filter->rc = send_input_flush(filter);
+        filter->flush = 0;
+    }
+
+    if (filter->eos) {
+        filter->rc = send_input_eos(filter);
+        filter->eos = 0;
     }
     
     return filter->rc;
@@ -827,7 +849,11 @@ static int modperl_filter_add_connection(conn_rec *c,
 
             if (handlers[i]->attrs & MP_FILTER_HAS_INIT_HANDLER &&
                 handlers[i]->next) {
-                int status = modperl_run_filter_init(f, handlers[i]->next);
+                int status = modperl_run_filter_init(
+                    f,
+                    (idx == MP_INPUT_FILTER_HANDLER
+                     ? MP_INPUT_FILTER_MODE : MP_OUTPUT_FILTER_MODE),
+                    handlers[i]->next);
                 if (status != OK) {
                     return status;
                 }
@@ -909,7 +935,11 @@ static int modperl_filter_add_request(request_rec *r,
 
             if (handlers[i]->attrs & MP_FILTER_HAS_INIT_HANDLER &&
                 handlers[i]->next) {
-                int status = modperl_run_filter_init(f, handlers[i]->next);
+                int status = modperl_run_filter_init(
+                    f,
+                    (idx == MP_INPUT_FILTER_HANDLER
+                     ? MP_INPUT_FILTER_MODE : MP_OUTPUT_FILTER_MODE),
+                    handlers[i]->next);
                 if (status != OK) {
                     return status;
                 }
@@ -968,6 +998,7 @@ void modperl_input_filter_add_request(request_rec *r)
 
 void modperl_filter_runtime_add(pTHX_ request_rec *r, conn_rec *c,
                                 const char *name,
+                                modperl_filter_mode_e mode,
                                 modperl_filter_add_t addfunc,
                                 SV *callback, const char *type)
 {
@@ -989,8 +1020,33 @@ void modperl_filter_runtime_add(pTHX_ request_rec *r, conn_rec *c,
             Perl_croak(aTHX_ "unable to resolve handler %s\n", handler->name);
         }
 
+        /* verify that the filter handler is of the right kind */
+        if (r == NULL) {
+            /* needs to have the FilterConnectionHandler attribute */
+            if (!(handler->attrs & MP_FILTER_CONNECTION_HANDLER)) {
+                Perl_croak(aTHX_ "Can't add connection filter handler '%s' "
+                           "since it doesn't have the "
+                           "FilterConnectionHandler attribute set",
+                           handler->name);
+            }
+        }
+        else {
+            /* needs to have the FilterRequestHandler attribute, but
+             * since by default request filters are not required to
+             * have the FilterRequestHandler attribute, croak only if
+             * some other attribute is set, but not
+             * FilterRequestHandler */
+            if (handler->attrs &&
+                !(handler->attrs & MP_FILTER_REQUEST_HANDLER)) {
+                Perl_croak(aTHX_ "Can't add request filter handler '%s' "
+                           "since it doesn't have the "
+                           "FilterRequestHandler attribute set",
+                           handler->name);
+            }
+        }
+
         if (handler->attrs & MP_FILTER_HAS_INIT_HANDLER && handler->next) {
-            int status = modperl_run_filter_init(f, handler->next);
+            int status = modperl_run_filter_init(f, mode, handler->next);
             if (status != OK) {
                 /* XXX */
             }
