@@ -26,7 +26,7 @@ static modperl_modglobal_key_t MP_modglobal_keys[] = {
     { NULL },
 };
 
-void modperl_modglobal_hash_keys(void)
+void modperl_modglobal_hash_keys(pTHX)
 {
     modperl_modglobal_key_t *gkey = MP_modglobal_keys;
 
@@ -314,7 +314,7 @@ typedef enum {
     MP_GLOBAL_GVHV,
     MP_GLOBAL_GVAV,
     MP_GLOBAL_GVIO,
-    MP_GLOBAL_SVPV,
+    MP_GLOBAL_SVPV
 } modperl_perl_global_types_e;
 
 typedef struct {
