@@ -289,6 +289,21 @@ char *modperl_server_desc(server_rec *s, apr_pool_t *p)
     return apr_psprintf(p, "%s:%u", s->server_hostname, s->port);
 }
 
+/* used in debug traces */
+MP_INLINE char *modperl_pid_tid(apr_pool_t *p)
+{
+    return apr_psprintf(p, "%lu"
+#if APR_HAS_THREADS
+                 "/%lu"
+#endif /* APR_HAS_THREADS */
+                 , (unsigned long)getpid()
+#if APR_HAS_THREADS
+                 , (unsigned long)apr_os_thread_current()
+#endif /* APR_HAS_THREADS */
+        );
+}
+
+    
 #define dl_librefs "DynaLoader::dl_librefs"
 #define dl_modules "DynaLoader::dl_modules"
 
@@ -668,3 +683,75 @@ void modperl_clear_symtab(pTHX_ HV *symtab)
         }
     }
 }
+
+#define SLURP_SUCCESS(action) \
+    if (rc != APR_SUCCESS) { \
+        SvREFCNT_dec(sv); \
+        Perl_croak(aTHX_ "Error " action " '%s': %s ", r->filename, \
+                   modperl_apr_strerror(rc)); \
+    }
+
+MP_INLINE SV *modperl_slurp_filename(pTHX_ request_rec *r, int tainted)
+{
+    SV *sv;
+    apr_status_t rc;
+    apr_size_t size;
+    apr_file_t *file;
+    
+    size = r->finfo.size;
+    sv = newSV(size);
+    file = r->finfo.filehand;
+    if (!file) {
+        rc = apr_file_open(&file, r->filename, APR_READ|APR_BINARY,
+                           APR_OS_DEFAULT, r->pool);
+        SLURP_SUCCESS("opening");
+    }
+
+    rc = apr_file_read(file, SvPVX(sv), &size);
+    SLURP_SUCCESS("reading");
+
+    MP_TRACE_o(MP_FUNC, "read %d bytes from '%s'\n", size, r->filename);
+    
+    if (r->finfo.size != size) {
+        SvREFCNT_dec(sv); 
+        Perl_croak(aTHX_ "Error: read %d bytes, expected %d ('%s')",
+                   size, r->finfo.size, r->filename);
+    }
+
+    rc = apr_file_close(file);
+    SLURP_SUCCESS("closing");
+    
+    SvPVX(sv)[size] = '\0';
+    SvCUR_set(sv, size);
+    SvPOK_on(sv);
+
+    if (tainted) {
+        SvTAINTED_on(sv);
+    }
+    else {
+        SvTAINTED_off(sv);
+    }
+    
+    return newRV_noinc(sv);
+}
+
+#ifdef MP_TRACE
+/* XXX: internal debug function */
+/* any non-false value for MOD_PERL_TRACE/PerlTrace enables this function */
+void modperl_apr_table_dump(pTHX_ apr_table_t *table, char *name)
+{
+    int i;
+    const apr_array_header_t *array;
+    apr_table_entry_t *elts;
+
+    array = apr_table_elts(table);
+    elts  = (apr_table_entry_t *)array->elts;
+    modperl_trace(MP_FUNC, "Contents of table %s", name);
+    for (i = 0; i < array->nelts; i++) {
+        if (!elts[i].key || !elts[i].val) {
+            continue;
+        }
+        modperl_trace(MP_FUNC, "%s => %s", elts[i].key, elts[i].val);
+    }    
+}
+#endif
