@@ -8,14 +8,20 @@ use Apache::Test;
 use APR::Table ();
 
 use Apache::Const -compile => 'OK';
+use APR::Const -compile => 'OVERLAP_TABLES_MERGE';
 
 my $filter_count;
 my $TABLE_SIZE = 20;
 
+use constant HAVE_APACHE_2_0_47 => have_min_apache_version('2.0.47');
+
 sub handler {
     my $r = shift;
 
-    plan $r, tests => 17;
+    my $tests = 19;
+    $tests += 2 if HAVE_APACHE_2_0_47;
+                                                                                                    
+    plan $r, tests => $tests;
 
     my $table = APR::Table::make($r->pool, $TABLE_SIZE);
 
@@ -98,6 +104,33 @@ sub handler {
         ok $filter_count == $TABLE_SIZE;
     }
 
+    # overlay and compress routines
+    my $base = APR::Table::make($r->pool, $TABLE_SIZE);
+    my $add = APR::Table::make($r->pool, $TABLE_SIZE);
+
+    $base->set(foo => 'one');
+    $base->add(foo => 'two');
+
+    $add->add(foo => 'three');
+    $add->add(bar => 'beer');
+
+    my $overlay = $base->overlay($add, $r->pool);
+
+    my @foo = $overlay->get('foo');
+    my @bar = $overlay->get('bar');
+
+    ok @foo == 3;
+    ok $bar[0] eq 'beer';
+
+    # BACK_COMPAT_MARKER: make back compat issues easy to find :)
+    if (HAVE_APACHE_2_0_47) {
+        $overlay->compress(APR::OVERLAP_TABLES_MERGE);
+
+        # $add first, then $base
+        ok $overlay->get('foo') eq 'three, one, two';
+        ok $overlay->get('bar') eq 'beer';
+    }
+
     Apache::OK;
 }
 
