@@ -3,6 +3,16 @@
 #include "mod_perl.h"
 #include "apr_perlio.h"
 
+/* XXX: modperl_perl_gensym() cannot be used outside of httpd */
+static SV *apr_perlio_gensym(pTHX_ char *pack)
+{
+    GV *gv = newGVgen(pack);
+    SV *rv = newRV((SV*)gv);
+    (void)hv_delete(gv_stashpv(pack, TRUE), 
+                    GvNAME(gv), GvNAMELEN(gv), G_DISCARD);
+    return rv;
+}
+
 #if defined(PERLIO_LAYERS) && defined(PERLIO_K_MULTIARG) /* 5.7.2+ */
 
 /**********************************************************************
@@ -89,20 +99,25 @@ static PerlIO *PerlIOAPR_open(pTHX_ PerlIO_funcs *self,
     
     st = PerlIOSelf(f, PerlIOAPR);
 
+    /* XXX: can't reuse a wrapper mp_xs_sv2_APR__Pool */
+    /* XXX: should probably add checks on pool validity in all other callbacks */
     sv = args[narg-1];
-    /* XXX: modperl_sv2pool cannot be used outside of httpd */
-    st->pool = modperl_sv2pool(aTHX_ sv);
-  
+    if (SvROK(sv) && (SvTYPE(SvRV(sv)) == SVt_PVMG)) {
+        st->pool = (apr_pool_t *)SvIV((SV*)SvRV(sv));
+    }
+    else {
+        Perl_croak(aTHX_ "argument is not a blessed reference "
+                   "(expecting an APR::Pool  derived object)");
+    }
+    
     rc = apr_file_open(&st->file, path, apr_flag, APR_OS_DEFAULT, st->pool);
 
-#ifdef PERLIO_APR_DEBUG
-    Perl_warn(aTHX_ "PerlIOAPR_open obj=0x%lx, file=0x%lx, name=%s, rc=%d\n",
-              (unsigned long)f, (unsigned long)st->file,
-              path ? path : "(UNKNOWN)", rc);
-#endif
+    MP_TRACE_o(MP_FUNC, "obj=0x%lx, file=0x%lx, name=%s, rc=%d",
+               (unsigned long)f, (unsigned long)st->file,
+               path ? path : "(UNKNOWN)", rc);
 
     if (rc != APR_SUCCESS) {
-        PerlIOBase(f)->flags |= PERLIO_F_ERROR;
+        PerlIO_pop(aTHX_ f);
         return NULL;
     }
 
@@ -128,14 +143,11 @@ static PerlIO *PerlIOAPR_dup(pTHX_ PerlIO *f, PerlIO *o,
         PerlIOAPR *ost = PerlIOSelf(o, PerlIOAPR);
 
         rc = apr_file_dup(&fst->file, ost->file, ost->pool);
-
-#ifdef PERLIO_APR_DEBUG
-        Perl_warn(aTHX_ "PerlIOAPR_dup obj=0x%lx, "
-                        "file=0x%lx => 0x%lx, rc=%d\n",
-                  (unsigned long)f,
-                  (unsigned long)ost->file,
-                  (unsigned long)fst->file, rc);
-#endif
+        
+        MP_TRACE_o(MP_FUNC, "obj=0x%lx, "
+                   "file=0x%lx => 0x%lx, rc=%d",
+                   (unsigned long)f, (unsigned long)ost->file,
+                   (unsigned long)fst->file, rc);
 
         if (rc == APR_SUCCESS) {
             fst->pool = ost->pool;
@@ -152,20 +164,24 @@ static SSize_t PerlIOAPR_read(pTHX_ PerlIO *f, void *vbuf, Size_t count)
     apr_status_t rc;
 
     rc = apr_file_read(st->file, vbuf, &count);
+
+#ifdef MP_TRACE
+    {
+        char *trace_buf = (char *)apr_pcalloc(st->pool,
+                                                   sizeof(char*)*count);
+        memcpy((void*)trace_buf, vbuf, count);
+        trace_buf[count] = '\0';
+        MP_TRACE_o(MP_FUNC, "count %d, [%s]", (int)count, (char*) trace_buf);
+    }
+#endif
+    
     if (rc == APR_EOF) {
         PerlIOBase(f)->flags |= PERLIO_F_EOF;
         return count;
     }
     else if (rc != APR_SUCCESS) {
-#ifdef PERLIO_APR_DEBUG
-        /* XXX: need to figure way to map APR errno to normal errno,
-         * so we can use SETERRNO to make the apr errors available to
-         * Perl's $!  */
         Perl_croak(aTHX_ "failed to read from file: %s",
                    modperl_apr_strerror(rc));
-#endif
-        PerlIOBase(f)->flags |= PERLIO_F_ERROR;
-        return -1;
     }
 
     return count;
@@ -176,10 +192,7 @@ static SSize_t PerlIOAPR_write(pTHX_ PerlIO *f, const void *vbuf, Size_t count)
     PerlIOAPR *st = PerlIOSelf(f, PerlIOAPR);
     apr_status_t rc;
 
-#if 0    
-     Perl_warn(aTHX_ "in write: count %d, %s\n",
-               (int)count, (char*) vbuf);
-#endif
+    MP_TRACE_o(MP_FUNC, "count %d, [%s]", (int)count, (char*) vbuf);
     
     rc = apr_file_write(st->file, vbuf, &count);
     if (rc == APR_SUCCESS) {
@@ -211,7 +224,7 @@ static IV PerlIOAPR_seek(pTHX_ PerlIO *f, Off_t offset, int whence)
     apr_status_t rc;
     apr_off_t seek_offset = 0;
 
-#if MP_LARGE_FILES_CONFLICT
+#ifdef MP_LARGE_FILES_CONFLICT
     if (offset != 0) {
         Perl_croak(aTHX_ "PerlIO::APR::seek with non-zero offset"
                    "is not supported with Perl built w/ -Duselargefiles"
@@ -266,7 +279,7 @@ static IV PerlIOAPR_close(pTHX_ PerlIO *f)
     IV code = PerlIOBase_close(aTHX_ f);
     apr_status_t rc;
 
-#ifdef PERLIO_APR_DEBUG
+#ifdef MP_TRACE
     const char *new_path = NULL;
     apr_os_file_t os_file;
 
@@ -280,9 +293,9 @@ static IV PerlIOAPR_close(pTHX_ PerlIO *f)
         Perl_croak(aTHX_ "filedes retrieval failed!");
     }
 
-    Perl_warn(aTHX_ "PerlIOAPR_close obj=0x%lx, file=0x%lx, fd=%d, name=%s\n",
-              (unsigned long)f, (unsigned long)st->file, os_file,
-              new_path ? new_path : "(UNKNOWN)");
+    MP_TRACE_o(MP_FUNC, "obj=0x%lx, file=0x%lx, fd=%d, name=%s",
+               (unsigned long)f, (unsigned long)st->file, os_file,
+               new_path ? new_path : "(UNKNOWN)");
 #endif
 
     if (PL_dirty) {
@@ -320,9 +333,7 @@ static IV PerlIOAPR_fill(pTHX_ PerlIO *f)
         PerlIO_get_base(f);  /* allocate via vtable */
     }
         
-#if 0
-     Perl_warn(aTHX_ "ask to fill %d chars\n", count);
-#endif
+    MP_TRACE_o(MP_FUNC, "asked to fill %d chars", count);
 
     rc = apr_file_read(st->file, st->base.ptr, &count);
     if (rc != APR_SUCCESS) {
@@ -330,9 +341,7 @@ static IV PerlIOAPR_fill(pTHX_ PerlIO *f)
         return -1;
     }
 
-#if 0    
-     Perl_warn(aTHX_ "got to fill %d chars\n", count);
-#endif
+    MP_TRACE_o(MP_FUNC, "got to fill %d chars", count);
 
     avail = count; /* apr_file_read() sets how many chars were read in count */
     if (avail <= 0) {
@@ -372,11 +381,17 @@ static IV PerlIOAPR_eof(pTHX_ PerlIO *f)
     return -1;
 }
 
+/* 5.8.0 doesn't export PerlIOBase_noop_fail, so we duplicate it here */
+static IV PerlIOAPR_noop_fail(pTHX_ PerlIO *f)
+{
+    return -1;
+}
+
 static PerlIO_funcs PerlIO_APR = {
     sizeof(PerlIO_funcs),
     "APR",
     sizeof(PerlIOAPR),
-    PERLIO_K_MULTIARG,
+    PERLIO_K_MULTIARG | PERLIO_K_RAW,
     PerlIOAPR_pushed,
     PerlIOBase_popped,
     PerlIOAPR_open,
@@ -391,7 +406,7 @@ static PerlIO_funcs PerlIO_APR = {
     PerlIOAPR_tell,
     PerlIOAPR_close,
     PerlIOAPR_flush,            /* flush */
-    PerlIOBase_noop_fail,       /* fill */
+    PerlIOAPR_noop_fail,        /* fill */
     PerlIOAPR_eof,
     PerlIOBase_error,
     PerlIOBase_clearerr,
@@ -445,7 +460,7 @@ PerlIO *apr_perlio_apr_file_to_PerlIO(pTHX_ apr_file_t *file, apr_pool_t *pool,
 
     st = PerlIOSelf(f, PerlIOAPR);
 
-#ifdef PERLIO_APR_DEBUG    
+#ifdef MP_TRACE
     {
         apr_status_t rc;
         apr_os_file_t os_file;
@@ -456,8 +471,8 @@ PerlIO *apr_perlio_apr_file_to_PerlIO(pTHX_ apr_file_t *file, apr_pool_t *pool,
             croak("filedes retrieval failed!");
         }
     
-        Perl_warn(aTHX_ "converting to PerlIO fd %d, mode '%s'\n",
-                  os_file, mode);
+        MP_TRACE_o(MP_FUNC, "converting to PerlIO fd %d, mode '%s'",
+                   os_file, mode);
     }
 #endif
 
@@ -470,8 +485,7 @@ PerlIO *apr_perlio_apr_file_to_PerlIO(pTHX_ apr_file_t *file, apr_pool_t *pool,
 
 static SV *apr_perlio_PerlIO_to_glob(pTHX_ PerlIO *pio, apr_perlio_hook_e type)
 {
-    /* XXX: modperl_perl_gensym() cannot be used outside of httpd */
-    SV *retval = modperl_perl_gensym(aTHX_ "APR::PerlIO"); 
+    SV *retval = apr_perlio_gensym(aTHX_ "APR::PerlIO"); 
     GV *gv = (GV*)SvRV(retval); 
 
     gv_IOadd(gv); 
@@ -502,7 +516,7 @@ SV *apr_perlio_apr_file_to_glob(pTHX_ apr_file_t *file, apr_pool_t *pool,
                                      type);
 }
 
-#elif !defined(PERLIO_LAYERS) && !defined(WIN32) /* NOT PERLIO_LAYERS (5.6.1) */
+#else /* NOT PERLIO_LAYERS (5.6.1) */
 
 static FILE *apr_perlio_apr_file_to_FILE(pTHX_ apr_file_t *file,
                                          apr_perlio_hook_e type)
@@ -528,14 +542,12 @@ static FILE *apr_perlio_apr_file_to_FILE(pTHX_ apr_file_t *file,
         Perl_croak(aTHX_ "filedes retrieval failed!");
     }
 
-#ifdef PERLIO_APR_DEBUG
-    Perl_warn(aTHX_ "converting fd %d\n", os_file);
-#endif
+    MP_TRACE_o(MP_FUNC, "converting fd %d", os_file);
 
     /* let's try without the dup, it seems to work fine:
 
        fd = PerlLIO_dup(os_file);
-       Perl_warn(aTHX_ "fd old: %d, new %d\n", os_file, fd);
+       MP_TRACE_o(MP_FUNC, "fd old: %d, new %d\n", os_file, fd);
        if (!(retval = PerlIO_fdopen(fd, mode))) { 
        ...
        }
@@ -560,7 +572,7 @@ SV *apr_perlio_apr_file_to_glob(pTHX_ apr_file_t *file, apr_pool_t *pool,
                                 apr_perlio_hook_e type)
 {
     /* XXX: modperl_perl_gensym() cannot be used outside of httpd */
-    SV *retval = modperl_perl_gensym(aTHX_ "APR::PerlIO"); 
+    SV *retval = apr_perlio_gensym(aTHX_ "APR::PerlIO"); 
     GV *gv = (GV*)SvRV(retval); 
 
     gv_IOadd(gv); 
@@ -586,12 +598,5 @@ void apr_perlio_init(pTHX)
     APR_REGISTER_OPTIONAL_FN(apr_perlio_apr_file_to_glob);
 }
 
-#else
-
-void apr_perlio_init(pTHX)
-{
-    Perl_croak(aTHX_ "APR::PerlIO not usable with this version of Perl");
-}
-
 #endif /* PERLIO_LAYERS */
 
