@@ -44,41 +44,7 @@ static MP_INLINE SV *mpxs_Apache__Module_get_config(pTHX_
                                                     server_rec *s,
                                                     ap_conf_vector_t *v)
 {
-    MP_dSCFG(s);
-    module *modp;
-    const char *name;
-    void *ptr;
-    PTR_TBL_t *table;
-    SV *obj;
-
-    if (!v) {
-        v = s->module_config;
-    }
-
-    if (SvROK(pmodule)) {
-        name = SvCLASS(pmodule);
-    }
-    else {
-        STRLEN n_a;
-        name = SvPV(pmodule, n_a);
-    }
-
-    if (!(scfg->modules &&
-          (modp = apr_hash_get(scfg->modules, name, APR_HASH_KEY_STRING)))) {
-        return &PL_sv_undef;
-    }
-
-    if (!(ptr = ap_get_module_config(v, modp))) {
-        return &PL_sv_undef;
-    }
-
-    if (!(table = modperl_module_config_table_get(aTHX_ FALSE))) {
-        return &PL_sv_undef;
-    }
-
-    if (!(obj = modperl_svptr_table_fetch(aTHX_ table, ptr))) {
-        return &PL_sv_undef;
-    }
+    SV *obj = modperl_module_config_get_obj(aTHX_ pmodule, s, v);
 
     return SvREFCNT_inc(obj);
 }
